import numpy as np
from dynasim.simulators import *
from dynasim.actuators import mdof_shaker, point_shakers
import scipy.integrate as integrate
import warnings

class state_space_system:
    '''
    Base class for any state space represented system
    '''

    def __init__(self) -> int:

        return 0

    def gen_state_matrices(self) -> int:
        '''
        Generate state matrices A, H, and nonlinear matrix An based on system parameters
        '''

        self.A = np.concatenate((
            np.concatenate((np.zeros((self.dofs,self.dofs)), np.eye(self.dofs)), axis=1),
            np.concatenate((-np.linalg.inv(self.M) @ self.K, -np.linalg.inv(self.M) @ self.C), axis=1)
        ), axis=0)

        self.H = np.concatenate((np.zeros((self.dofs, self.dofs)), np.linalg.inv(self.M)), axis=0)

        match [self.Cn, self.Kn]:
            case [None,None]:
                self.An = None
            case [_,None]:
                self.An = np.concatenate((
                    np.zeros((self.dofs, 2*self.dofs)),
                    np.concatenate((np.zeros_like(self.Cn), -np.linalg.inv(self.M) @ self.Cn), axis=1)
                ), axis=0)
            case [None, _]:
                self.An = np.concatenate((
                    np.zeros((self.dofs, 2*self.dofs)),
                    np.concatenate((-np.linalg.inv(self.M) @ self.Kn, np.zeros_like(self.Kn)), axis=1)
                ), axis=0)
            case [_,_]:
                self.An = np.concatenate((
                    np.zeros((self.dofs, 2*self.dofs)),
                    np.concatenate((-np.linalg.inv(self.M) @ self.Kn, -np.linalg.inv(self.M) @ self.Cn), axis=1)
                ), axis=0)
        
        return 0
    
    def gen_obs_matrices(self) -> int:
        '''
        Generate observation matrices B and D based on system parameters
        '''
        
        self.B = np.concatenate((-np.linalg.inv(self.M) @ self.K, -np.linalg.inv(self.M) @ self.C), axis=1)
        self.D = np.linalg.inv(self.M)

        match [self.Cn, self.Kn]:
            case [None,None]:
                self.Bn = None
            case [_, None]:
                self.Bn = np.concatenate((np.zeros_like(self.Cn), -np.linalg.inv(self.M) @ self.Cn), axis=1)
            case [None, _]:
                self.Bn = np.concatenate((-np.linalg.inv(self.M) @ self.Kn, np.zeros_like(self.Kn)), axis=1)
            case [_,_]:
                self.Bn = np.concatenate((-np.linalg.inv(self.M) @ self.Kn, -np.linalg.inv(self.M) @ self.Cn), axis=1)

        return 0

class cont_ss_system(state_space_system):
    '''
    Base class for continuous system represented in the state space
    '''

    def __init__(self, M=None, C=None, K=None, modes=1):

        self.M = M
        self.C = C
        self.K = K
        self.Cn = None  # no nonlinearities for now
        self.Kn = None  # no nonlinearities for now
        self.dofs = modes

        self.gen_state_matrices
    
    def simulate(self, tt, z0=None, simulator=None):
        '''
        Simulate the system for a given time using the specified simulator.

        Args:
            tt: The vector of time samples
            z0: The initial state of the system. Defaults to None.
            simulator: The simulator to use for the simulation. Defaults to scipy.solve_ivp.

        Returns:
            A dictionary containing the system's response displacement and velocity over time.
        '''

        # instantiate time
        self.t = tt

        if hasattr(self, 'excitations'):
            if self.excitations is not None:
                # create shaker object
                self.shaker = point_shakers(self.excitations, self.xx)
                # generate forcing series
                # self.f = self.shaker.generate(tt)
                ff = self.shaker.generate(tt)
                pp = np.zeros((self.n_modes, tt.shape[0]))
                for n in range(self.n_modes):
                    pp[n,:] = integrate.simpson(self.phi_n[:,n].reshape(-1,1) * ff, self.xx, axis=0).reshape(-1)
                self.f = pp
            else:
                self.f = None
        else:
            self.f = None

        # initiate simulator
        if simulator is None:
            # self.simulator = scipy_ivp(self)
            self.simulator = rk4(self)
        else:
            self.simulator = simulator(self)

        # initial conditions
        if z0 is None:
            # warnings.warn('No initial conditions provided, proceeding with zero initial state', UserWarning)
            tau0 = np.zeros((2*self.dofs))
            if all([e is None for e in self.excitations]):
                warnings.warn('Zero initial condition and zero excitations, what do you want??', UserWarning)
        else:
            tau0 = z0
        
        # simulate
        qq = self.simulator.sim(tt, tau0)
        return qq

class mdof_system(state_space_system):
    '''
    Base class for generic mdof system
    '''

    def __init__(self, M=None, C=None, K=None, Cn=None, Kn=None, nonlinearity=None):

        self.M = M
        self.C = C
        self.K = K
        self.Cn = Cn
        self.Kn = Kn
        self.dofs = M.shape[0]
        self.nonlinearity = nonlinearity

        self.gen_state_matrices()
        self.gen_obs_matrices()
    
    def nonlin_transform(self, z):

        if self.nonlinearity is not None:

            x_ = z[:self.dofs] - np.concatenate((np.zeros_like(z[:1]), z[:self.dofs-1]))
            x_dot = z[self.dofs:] - np.concatenate((np.zeros_like(z[:1]), z[self.dofs:-1]))

            return np.concatenate((
                self.nonlinearity.gk_func(x_, x_dot),
                self.nonlinearity.gc_func(x_, x_dot)
            ))
        
        else:
            return np.zeros_like(z)

    def simulate(self, tt, z0=None, simulator=None):
        '''
        Simulate the system for a given time using the specified simulator.

        Args:
            tt: The vector of time samples
            z0: The initial state of the system. Defaults to None.
            simulator: The simulator to use for the simulation. Defaults to scipy.solve_ivp.

        Returns:
            A dictionary containing the system's response displacement and velocity over time.
        '''

        try:
            # instantiate time
            self.t = tt

            if hasattr(self, 'excitations'):
                # create shaker object
                self.shaker = mdof_shaker(self.excitations)
                # generate forcing series
                self.f = self.shaker.generate(tt)
            else:
                self.f = None

            # initiate simulator
            if simulator is None:
                # self.simulator = scipy_ivp(self)
                self.simulator = rk4(self)
            else:
                self.simulator = simulator(self)

            # initial conditions
            if z0 is None:
                warnings.warn('No initial conditions provided, proceeding with zero initial state', UserWarning)
                z0 = np.zeros((2*self.dofs))
                if all([e is None for e in self.excitations]):
                    warnings.warn('Zero initial condition and zero excitations, what do you want??', UserWarning)
            
            # simulate
            simulated_state_data = self.simulator.sim(tt, z0)
            
            # generate acceleration observation data
            z = simulated_state_data['z']
            if self.Bn is None:
                acceleration = self.B @ z + self.D @ self.f
            else:
                acceleration = self.B @ z + self.Bn @ self.nonlin_transform(z) + self.D @ self.f
                
            simulated_state_data.update({'acc': acceleration})
            
            # Final debug print to check the updated dictionary
            # print("Updated simulated_state_data:", simulated_state_data)
            
        except Exception as e:
            print("An error occurred during simulation:", e)
            simulated_state_data = None
            
        return simulated_state_data





        


import numpy as np
import warnings

class nonlinearity:

    def __init__(self, dofs):

        self.dofs = dofs
        if self.dofs is not None:
            self.Cn = np.zeros((self.dofs, self.dofs))
            self.Kn = np.zeros((self.dofs, self.dofs))

    def gc_func(self, x, xdot):
        return np.zeros_like(xdot)
    
    def gk_func(self, x, xdot):
        return np.zeros_like(x)

class grid_exponent_stiffness(nonlinearity):
    
    def __init__(self, kn_h, kn_v, exponent=3, arr_shape=None):
        self.exponent = exponent
        match kn_h:
            case np.ndarray():
                self.kn_h = kn_h
                arr_shape = kn_h.shape
            case None:
                warnings.warn('No horizontal nonlinear stiffness parameters provided, proceeding with zero', UserWarning)
                self.kn_h = None
                arr_shape = None
            case _:
                if arr_shape is None:
                    warnings.warn('Under defined nonlinearity, proceeding with zero nonlinearity')
                    self.kn_h = None
                else:
                    self.kn_h = kn_h * np.ones(arr_shape)
        match kn_v:
            case np.ndarray():
                self.kn_v = kn_v
            case None:
                warnings.warn('No vertical nonlinear stiffness parameters provided, proceeding with zero', UserWarning)
                self.kn_v = None
            case _:
                if arr_shape is None:
                    warnings.warn('Under defined nonlinearity, proceeding with zero nonlinearity')
                    self.kn_v = None
                else:
                    self.kn_v = kn_v * np.ones(arr_shape)
        dofs = 2 * arr_shape[0] * arr_shape[1]
        m, n = arr_shape
        super().__init__(dofs)
        
        Kn = np.zeros((dofs, dofs))
        
        def index(i, j, direction):
            return 2 * (i * n + j) + direction # 0 for horizontal, 1 for vertical
        
        for i in range(m):
            for j in range(n):
                x_index = index(i, j, 0)
                y_index = index(i, j, 1)
                
                Kn[x_index, x_index] = self.kn_h[i, j]
                Kn[y_index, y_index] = self.kn_v[i, j]
                
                if (0 < j):
                    Kn[index(i, j-1, 0), x_index] = -self.kn_h[i, j]
                if (0 < i):
                    Kn[index(i-1, j, 1), y_index] = -self.kn_v[i, j]
                    
        self.Kn = Kn
    
    def gk_func(self, x, xdot):
        return np.sign(x) * np.abs(x)**self.exponent

class exponent_stiffness(nonlinearity):

    def __init__(self, kn_, exponent=3, dofs=None):
        self.exponent = exponent
        match kn_:
            case np.ndarray():
                self.kn_ = kn_
                dofs = kn_.shape[0]
            case None:
                warnings.warn('No nonlinear stiffness parameters provided, proceeding with zero', UserWarning)
                self.kn_ = None
                dofs = None
            case _:
                if dofs is None:
                    warnings.warn('Under defined nonlinearity, proceeding with zero nonlinearity')
                    self.kn_ = None
                else:
                    self.kn_ = kn_ * np.ones(dofs)
        super().__init__(dofs)

        self.Kn = np.diag(kn_) - np.diag(kn_[1:], 1)
        
    def gk_func(self, x, xdot):
        return np.sign(x) * np.abs(x)**self.exponent

class exponent_damping(nonlinearity):

    def __init__(self, cn_, exponent=0.5, dofs=None):
        self.exponent = exponent
        match cn_:
            case np.ndarray():
                self.cn_ = cn_
                dofs = cn_.shape[0]
            case None:
                warnings.warn('No nonlinear damping parameters provided, proceeding with zero', UserWarning)
                self.cn_ = None
                dofs = None
            case _:
                if dofs is None:
                    warnings.warn('Under defined nonlinearity, proceeding with zero nonlinearity')
                    self.cn_ = None
                else:
                    self.cn_ = cn_ * np.ones(dofs)
        super().__init__(dofs)
        
        self.Cn = np.diag(cn_) - np.diag(cn_[1:], 1)
    
    def gc_func(self, x, xdot):
        return np.sign(xdot) * np.abs(xdot)**self.exponent
    
class vanDerPol(nonlinearity):

    def __init__(self, cn_, dofs=None):
        match cn_:
            case np.ndarray():
                self.cn_ = cn_
                dofs = cn_.shape[0]
            case None:
                warnings.warn('No nonlinear damping parameters provided, proceeding with zero', UserWarning)
                self.cn_ = None
                dofs = None
            case _:
                if dofs is None:
                    warnings.warn('Under defined nonlinearity, proceeding with zero nonlinearity')
                    self.cn_ = None
                else:
                    self.cn_ = cn_ * np.ones(dofs)
        super().__init__(dofs)

        self.Cn = np.diag(cn_) - np.diag(cn_[1:], 1)

    def gc_func(self, x, xdot):
        return (x**2 - 1) * xdot
    
    




import numpy as np
import scipy
import scipy.signal
from math import pi



class excitation():

    def generate(self, time):
        return self._generate(time)
    
class impulse(excitation):
    '''
    Impulse excitation
    '''
    
    def __init__(self, imp_times = [0], width = 0.1, f0=0.1):
        self.imp_times = imp_times
        self.width = width
        self.f0 = f0
        
    def _generate(self, time, seed=43810):
        ns = time.shape[0]
        nt_width = int(self.width/(time[1]-time[0])) // 2
        f = np.zeros(ns)
        imp_locs = [np.argmin(np.abs(time - imp_time)) for imp_time in self.imp_times]
        for imp_loc in imp_locs:
            f[imp_loc - nt_width:imp_loc + nt_width] = self.f0
        return f
    
class sinusoid(excitation):
    '''
    Single sinusoidal signal with central frequency w, amplitude f0, and phase phi
    '''

    def __init__(self, w, f0=1.0, phi=0):
        self.w = w
        self.f0 = f0
        self.phi = phi

    def _generate(self, time, seed=43810):
        return self.f0 * np.sin(self.w*time + self.phi)
    
class white_gaussian(excitation):
    '''
    White Gaussian noise with variance f0, and mean
    '''

    def __init__(self, f0, mean=0.0):
        self.f0 = f0
        self.u = mean
    
    def _generate(self, time, seed=43810):
        ns = time.shape[0]
        np.random.seed(seed)
        return np.random.normal(self.u, self.f0*np.ones((ns)))
    
class banded_noise(excitation):

    def __init__(self, bandwidth, amplitude):
        self.bandwidth = bandwidth
        self.amplitude = amplitude

    def fftnoise(self, f):
        f = np.array(f, dtype='complex')
        Np = (len(f) - 1) // 2
        phases = np.random.rand(Np) * 2 * np.pi
        phases = np.cos(phases) + 1j * np.sin(phases)
        f[1:Np+1] *= phases
        f[-1:-1-Np:-1] = np.conj(f[1:Np+1])
        return np.fft.ifft(f).real

    def band_limited_noise(self, min_freq, max_freq, samples=1024, samplerate=1):
        freqs = np.abs(np.fft.fftfreq(samples, 1/samplerate))
        f = np.zeros(samples)
        idx = np.where(np.logical_and(freqs>=min_freq, freqs<=max_freq))[0]
        f[idx] = 1
        return self.fftnoise(f)

    def _generate(self, time, seed=43810):
        np.random.seed(seed)
        return self.amplitude * self.band_limited_noise(self.bandwidth[0], self.bandwidth[1], samples=len(time), samplerate=1/(time[1]-time[0]))
    
class sine_sweep(excitation):
    '''
    Sine sweep signal
    '''

    def __init__(self, w_l, w_u, F0=1.0, scale='linear'):
        self.w_l = w_l
        self.w_u = w_u
        self.F0 = F0
        self.scale = scale

    def _generate(self, time, seed=43810):
        f0 = self.w_l / (2*pi)
        f1 = self.w_u / (2*pi)
        F =  self.F0 * scipy.signal.chirp(time, f0, time[-1], f1, method=self.scale)
        return F
    
class rand_phase_ms(excitation):
    '''
    Random-phase multi-sine
    '''

    def __init__(self, freqs, Sx):
        self.freqs = freqs
        self.Sx = Sx

    def _generate(self, time, seed=43810):

        np.random.seed(seed)
        phases = np.random.randn(self.freqs.shape[0]) * pi/2
        F_mat = np.sin(time.reshape(-1,1) @ self.freqs.reshape(1,-1) + phases.T)

        return (F_mat @ self.Sx).reshape(-1)

class mdof_shaker():
    '''
    MDOF shaker class generates force signals at each DOF using excitation class
    '''

    def __init__(self, excitations=None, seed=43810):

        self.excitations = excitations
        self.dofs = len(excitations)
        self.seed = seed

    def generate(self, time):
        nt = time.shape[0]
        self.f = np.zeros((self.dofs,nt))
        for n, excite in enumerate(self.excitations):
            match excite:
                case excitation():
                    self.f[n, :] = self.excitations[n]._generate(time, self.seed+n)
                case np.ndarray():
                    self.f[n, :] = self.excitations[n]
                case None:
                    self.f[n, :] = np.zeros(nt)

        return self.f
    
class point_shakers():
    '''
    Point shakers class generates force signals at a specific locations using excitation class
    '''

    def __init__(self, excitations=None, xx=None, seed=43810):

        self.excitations = excitations
        "excitations is a list of dictionaries containing location and excitation"
        self.xx = xx
        self.seed = seed

    def generate(self, time):
        nt = time.shape[0]
        nx = self.xx.shape[0]
        self.f = np.zeros((nx, nt))
        for excite in self.excitations:
            self.loc_id = np.argmin(np.abs(self.xx-excite['loc']))
            match excite['excitation']:
                case excitation():
                    self.f[self.loc_id, :] = excite['excitation']._generate(time, self.seed)
                case np.ndarray():
                    self.f[self.loc_id, :] = self.excitation
        return self.f
        


